@@ -4,7 +4,8 @@
 #  ___________________________________________________________________________
 #
 #  Pyomo: Python Optimization Modeling Objects
-#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
+#  Copyright (c) 2008-2022
+#  National Technology and Engineering Solutions of Sandia, LLC
 #  Under the terms of Contract DE-NA0003525 with National Technology and
 #  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
 #  rights in this software.
@@ -15,15 +16,23 @@ from __future__ import division
 
 import logging
 
-import pyutilib.th as unittest
-from six import StringIO
+import pyomo.common.unittest as unittest
+from io import StringIO
 
-from pyomo.common.dependencies import networkx_available
+from pyomo.common.dependencies import networkx_available, matplotlib_available
 from pyomo.common.log import LoggingIntercept
-from pyomo.environ import ConcreteModel, Constraint, Objective, Var, Integers, minimize, RangeSet, Block, ConstraintList
-from pyomo.contrib.community_detection.detection import detect_communities, CommunityMap, \
-    community_louvain_available, matplotlib_available, community_louvain, matplotlib
-from pyomo.contrib.community_detection.community_graph import generate_model_graph
+from pyomo.common.tempfiles import TempfileManager
+from pyomo.environ import (
+    ConcreteModel, Constraint, Objective, Var, Integers, minimize,
+    RangeSet, Block, ConstraintList,
+)
+from pyomo.contrib.community_detection.detection import (
+    detect_communities, CommunityMap,
+    community_louvain_available, community_louvain
+)
+from pyomo.contrib.community_detection.community_graph import (
+    generate_model_graph,
+)
 
 from pyomo.solvers.tests.models.LP_unbounded import LP_unbounded
 from pyomo.solvers.tests.models.QP_simple import QP_simple
@@ -617,7 +626,10 @@ class TestDecomposition(unittest.TestCase):
         model = decode_model_1()
         community_map_object = detect_communities(model)
 
-        fig, pos = community_map_object.visualize_model_graph(filename='test_visualize_model_graph_1')
+        with TempfileManager:
+            fig, pos = community_map_object.visualize_model_graph(
+                filename=TempfileManager.create_tempfile(
+                    'test_visualize_model_graph_1.png'))
         correct_pos_dict_length = 5
 
         self.assertTrue(isinstance(pos, dict))
@@ -628,8 +640,11 @@ class TestDecomposition(unittest.TestCase):
         model = decode_model_2()
         community_map_object = detect_communities(model)
 
-        fig, pos = community_map_object.visualize_model_graph(type_of_graph='bipartite',
-                                                              filename='test_visualize_model_graph_2')
+        with TempfileManager:
+            fig, pos = community_map_object.visualize_model_graph(
+                type_of_graph='bipartite',
+                filename=TempfileManager.create_tempfile(
+                    'test_visualize_model_graph_2.png'))
         correct_pos_dict_length = 13
 
         self.assertTrue(isinstance(pos, dict))
@@ -642,8 +657,8 @@ class TestDecomposition(unittest.TestCase):
 
         community_map_object = cmo = detect_communities(model, random_seed=5)
         correct_partition = {3: 0, 4: 1, 5: 0, 6: 0, 7: 1, 8: 0}
-        correct_components = {'b[0].B[2].c', 'b[0].c2[1]', 'b[0].c1[3]', 'equality_constraint_list[1]',
-                              'b[1].c2[2]', 'b[1].x', 'b[0].x', 'b[0].y', 'b[0].z', 'b[0].obj[2]', 'b[1].c1[2]'}
+        correct_components = {"b[0].'B[2].c'", "b[0].'c2[1]'", "b[0].'c1[3]'", 'equality_constraint_list[1]',
+                              "b[1].'c2[2]'", 'b[1].x', 'b[0].x', 'b[0].y', 'b[0].z', "b[0].'obj[2]'", "b[1].'c1[2]'"}
 
         structured_model = cmo.generate_structured_model()
         self.assertIsInstance(structured_model, Block)
