@@ -1,9 +1,10 @@
 #  ___________________________________________________________________________
 #
 #  Pyomo: Python Optimization Modeling Objects
-#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
-#  Under the terms of Contract DE-NA0003525 with National Technology and 
-#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
+#  Copyright (c) 2008-2022
+#  National Technology and Engineering Solutions of Sandia, LLC
+#  Under the terms of Contract DE-NA0003525 with National Technology and
+#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
 #  rights in this software.
 #  This software is distributed under the 3-clause BSD License.
 #  ___________________________________________________________________________
@@ -158,7 +159,10 @@ class ScaleModel(Transformation):
                     v.setub(temp)
 
                 if v.value is not None:
-                    v.value = value(v) * scaling_factor
+                    # Since the value was OK in the unscaled space, it
+                    # should be safe to assume it is still valid in the
+                    # scaled space)
+                    v.set_value(value(v) * scaling_factor, skip_validation=True)
 
         # scale the objectives/constraints and perform the scaled variable substitution
         scale_constraint_dual = False
@@ -188,20 +192,25 @@ class ScaleModel(Transformation):
                                                remove_named_expressions=True)
 
                     # scale the rhs
-                    if c._lower is not None:
-                        c._lower = c._lower * scaling_factor
-                    if c._upper is not None:
-                        c._upper = c._upper * scaling_factor
+                    lower = c.lower
+                    upper = c.upper
+                    if lower is not None:
+                        lower = lower * scaling_factor
+                    if upper is not None:
+                        upper = upper * scaling_factor
 
                     if scaling_factor < 0:
-                        c._lower, c._upper = c._upper, c._lower
+                        lower, upper = upper, lower
 
                     if scale_constraint_dual and c in model.dual:
                         dual_value = model.dual[c]
                         if dual_value is not None:
                             model.dual[c] = dual_value / scaling_factor
-
-                    c.set_value((c._lower, body, c._upper))
+                    
+                    if c.equality:
+                        c.set_value((lower, body))
+                    else:
+                        c.set_value((lower, body, upper))
 
                 elif isinstance(c, _ObjectiveData):
                     c.expr = scaling_factor * \
@@ -263,7 +272,10 @@ class ScaleModel(Transformation):
             original_v = original_model.find_component(original_v_path)
 
             for k in scaled_v:
-                original_v[k].value = value(scaled_v[k]) / component_scaling_factor_map[scaled_v[k]]
+                original_v[k].set_value(
+                    value(scaled_v[k]) / component_scaling_factor_map[scaled_v[k]],
+                    skip_validation=True,
+                )
                 if check_reduced_costs and scaled_v[k] in scaled_model.rc:
                     original_model.rc[original_v[k]] = scaled_model.rc[scaled_v[k]] * component_scaling_factor_map[
                         scaled_v[k]] / objective_scaling_factor
